import time
import torch
import torch.nn as nn

from genn.generator import Generator
from genn.preprocessing import Preprocessing
from genn.controllers import *

class LSTMGenerator(Generator):
    """
        LSTM model for text generation.
        Parameters
        ----------
        preproObj : Preprocessing class object
            An object created using the class Preprocessing.

        nLayers: int
            The number of LSTM layers.
        
        batchSize: int
            The number of training instances per iteration.

        embSize: int
            The size of the vector in which each token will be embedded.
        
        lstmSize: int
            The number of features in the hidden state.
        
        epochs: int
            The number of cycles over the training data.
        
        dropout: float, optional, default 0
            The percentage of units that will be dropped per layer.
        
        embeddingOption: string, optional, default random
            The type of embeddings that will be used for the tokens.
            It can be glove, fasttext or random. Random means the NN.Embedding will be used.
        
        predictionIteration: int, optional, default 10
            The number of iterations for the prediction loop unless the model generates the stop token.
        
        glovePath: string, optional, default None
            The path for the GloVE embeddings if embeddingType is set to 'glove'.
        
        fastTextParams: dict, optional, default {}
            The parameters will be used while fine-tuning fastText to create the embeddings.
            Emtpy dictionary means the default training parameters will be used.
        
        loadFastText: string, optional, default None
            A path to the pretrained fastText model file.
        
        saveFastText: string, optional, default None
            A path to save the fastText model after training.
        
        fineTuneEmbs: bool, optional,  default False
            Whether to fine-tune the pretrained embeddings or not.

        selectionParams: dict, optional, default {"sType": 'topk', 'k': 5}
            sType: The type of selection method to use during prediciton. It is either 'topk' or 'nucleus', 't' and 'n' can also be used.
            k: in case of topk, it represents the k value.
            probThreshold: When using Nucleus selection, this represnts the probability threshold p. 

    """
    def __init__(self, preproObj, nLayers, batchSize, embSize, lstmSize,
                   epochs, dropout = 0, embeddingOption ='random',
                   genIteration = 10, glovePath = None, fastTextParams = {},
                   loadFastText = None, saveFastText = None, fineTuneEmbs = False, 
                   selectionParams = {"sType": 'topk', 'k': 5}):

        super(LSTMGenerator, self).__init__(preproObj, nLayers, batchSize, embSize,
                                        lstmSize, epochs, dropout, embeddingOption, genIteration,
                                        glovePath, fastTextParams, loadFastText, saveFastText, fineTuneEmbs,
                                        selectionParams)

        self.lstm = nn.LSTM(self.embSize,
                        self.rnnSize,
                        num_layers = self.nLayers,
                        dropout = self.dropout,
                        batch_first = True)
        self.dense = nn.Linear(self.rnnSize, self.nVocab)

    def forward(self, x, length, prevState):
        """
            params:
                x: the tensor of input instances.
                length: the length that will be used from each training instance to handle paddings.
                prev_state: the hidden and the cell states

        """

        embed = self.embedding(x)
        packedInput = self.packSrc(embed, length)
        packedOutput, state = self.lstm(packedInput, prevState)
        padded, _ = self.padPack(packedOutput) 
        logits = self.dense(padded)
        return logits, state

    def run(self):
        """
            Training method
        """
        self.to(self.device)
        criterion, optimizer = self.getLossAndTrainOp(0.01)

        batchCount = 0
        batchesLen = (len(self.preproObj) // self.batchSize) * self.epochs

        for e in range(self.epochs):        
            batches = self.getBatches()
            
            stateH, stateC = self.zeroState(2, self.batchSize)
            stateH = stateH.to(self.device)
            stateC = stateC.to(self.device)
            
            batchTimes = []
            for batch in batches:

                startTime = time.time()
                srcLengths, _ ,srcBatch,trgBatch = self.getSrcTrg(batch)
                
                batchCount += 1

                self.train()

                optimizer.zero_grad()
                
                trgBatch = torch.t(trgBatch).to(self.device)
                srcBatch = srcBatch.to(self.device)
                
                logits, (stateH, stateC) = self(srcBatch,srcLengths,(stateH,stateC))

                loss = criterion(logits.transpose(1, 2), trgBatch)

                stateH = stateH.detach()
                stateC = stateC.detach()

                lossValue = loss.item()

                loss.backward(retain_graph=True)

                optimizer.step()

                loss.backward()

                _ = nn.utils.clip_grad_norm_(
                self.parameters(), 5)

                optimizer.step()

                batchTime = time.time() - startTime
                batchTimes.append(batchTime)

                if batchCount % 10 == 0:
                    timeChecker(batchTimes, batchCount, batchesLen, e, self.epochs, lossValue)



import csv, json, re 
from torchtext.data import Dataset, Field
from genn.importers import *
from genn.controllers import *
from genn.Document import Document
from collections import Counter
import numpy as np



class Preprocessing(Dataset):
    
    __tokPattern = r"""[0-9A-Za-z_]*[A-Za-z_-]+[0-9A-Za-z_]*|\.|\!|\?|\d+|\-|%|[.,!?;'"]"""
    __supportedExtensions = ['txt', 'csv', 'json']
    __seedAttrs = ['nFirst', 'minFreq']


    def __init__(self,
                    fileParams = {}, tokenizationOption='regex',
                    seedParams = {'nFirst': 1, 'minFreq': 5},
                    fieldParams = {'lower': True, 'eos_token': '<!EOS!>'},spacyObj = None):

        self.__fileName, self.__fileExtension, self.__parsingColumn = checkFileParams(fileParams)
        self.__seedParams = checkSeedParams(seedParams)
        self.__DataVocab = Field(**fieldParams)
        self.__spacyObj = spacyObj
        self.__customTokenize = self.__tokenizationMethod(tokenizationOption)
        self.__readFile()
    
    @property
    def getFileName(self):
        return self.__fileName
    
    @property
    def getVocab(self):
        return self.__DataVocab

    def __readFile(self):
        text = readFiles(self.__fileName, self.__fileExtension, self.__parsingColumn)                
        self.examples = self.__getObjects(text)
        self.__seeds = getStartWords(self.__seedParams, text)
        self.__build_vocab()
    
    def __getObjects(self, text):   
        self.fields = {"src": self.__DataVocab}     
        return [Document(**self.__tokenize(instance)) 
                for instance in text]

    def __build_vocab(self):
        self.__DataVocab.build_vocab(self)
        for instance in self.examples:
            instance.create_tokens(self.__DataVocab) 

    def __regexTokenization(self, document):
        return re.findall(self.__tokPattern, document)

    def __nltkTokenization(self, document):
        return self.tokenizer(document)

    def __spacyTokenization(self,instance):
        return [entity.text.strip() for entity in self.__spacyObj(instance) if entity.text.strip() ]

    def __tokenize(self,instance):
        instance = self.__customTokenize(instance)
        return {'src': instance, 'trg': instance[1: ]}
    
    @checkParams(str)
    def __tokenizationMethod(self, param):
        param = param.lower()

        if param == 'nltk':
            self.tokenizer = importNltk()
            return self.__nltkTokenization

        elif param == 'regex':
            return self.__regexTokenization
        
        elif param == 'spacy':
            if not self.__spacyObj:
                raise Exception("Please provide the spacy object to tokenize with.")
                
            return self.__spacyTokenization
                
        
        raise Exception("The parameter 'tokenizationOption' can only be nltk, regex and spacy")

    def getSeed(self):
        """
            return a weighted seed. 
            In case static seed is enabled, then the most frequent token will be the seed.
        """
        seeds = list(self.__seeds.keys())
        probs = list(self.__seeds.values())
        return np.random.choice(seeds, 1 , probs).tolist()


            
            








from genn import importers
from genn import controllers
from genn.lstm import LSTMGenerator
from genn.gru import GRUGenerator
from genn.Document import Document
from genn.preprocessing import Preprocessing


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="abdoTheBest",
    version="0.7.7",
    author="Abdelrahman Mahmoud, Fahed Sabellioglu",
    author_email="magedmahmoud@std.sehir.edu.tr, fahedshaabani@std.sehir.edu.tr",
    description="High level interface for text applications using PyTroch RNN's.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/FahedSabellioglu/genn',
    packages=setuptools.find_packages(),
    install_requires  = ['torch==1.4.0','numpy', 'fasttext'],
    license = 'MIT'
)


import time

import torch
import torch.nn as nn

from genn.controllers import *
from genn.generator import Generator
from genn.preprocessing import Preprocessing


class GRUGenerator(Generator):
	"""
	GRU arch for text generation

	Parameters
	----------
	datasetObj : Preprocessing class object
		An object created using the class Preprocessing.

	nLayers: int
		The number of GRU layers

	batchSize: int
		The number of training instances per iteration.

	embSize: int
		The size of the vector in which each token will be embedded.

	gruSize: int
		The number of features in the hidden state.

	epochs: int
		The number of cycles over the training data

	dropout: float, optional, default 0
		The percentage of units that will be dropped per layer.

	embeddingOption: string, optional, default random
		The type of embeddings that will be used for the tokens.
		It can be glove, fasttext or random. Random means the NN.Embedding will be used.

	predictionIteration: int, optional, default 10
		The number of iterations for the prediction loop unless the model generates the stop token.

	glovePath: string, optional, default None
		The path for the GloVE embeddings if embeddingType is set to 'glove'.

	fastTextParams: dict, optional, default {}
		The parameters will be used while fine-tuning fastText to create the embeddings.
		Emtpy dictionary means the default training parameters will be used.

	loadFastText: string, optional, default None
		A path to the pretrained fastText model file.

	saveFastText: string, optional, default None
		A path to save the fastText model after training.

	fineTuneEmbs: bool, optional,  default False
		Whether to fine-tune the pretrained embeddings or not.

	selectionParams: dict, optional, default {"sType": 'topk', 'k': 5}
		sType: The type of selection method to use during prediciton.
			It is either 'topk' or 'nucleus', 't' and 'n' can also be used.
		k: in case of topk, it represents the k value.
		probThreshold: When using Nucleus selection, this represnts the probability threshold p.
	"""

	def __init__(
				self,preproObj, nLayers, batchSize, embSize, gruSize,epochs,
				dropout = 0.0 , embeddingOption ='fasttext',genIteration = 10, glovePath = None,
				fastTextParams = {},loadFastText = None, saveFastText = None, fineTuneEmbs = False,
				selectionParams = {"sType": 'topk', 'k': 5, 'probThreshold': 0.5}):


		super(GRUGenerator, self).__init__(preproObj, nLayers, batchSize, embSize,
										gruSize, epochs, dropout, embeddingOption, genIteration,
										glovePath, fastTextParams, loadFastText, saveFastText, fineTuneEmbs,
										selectionParams)

		self.gru = nn.GRU(self.embSize,
						self.rnnSize,
						num_layers = self.nLayers,
						dropout = self.dropout,
						batch_first = True)
		self.dense = nn.Linear(self.rnnSize, self.nVocab)

	def forward(self, x, length, prevState):
		"""
			params:
				x: the tensor of input instances.
				length: the length that will be used from each training instance to handle paddings.
				prev_state: cell state.

		"""
		embed = self.embedding(x)
		packedInput = self.packSrc(embed, length)
		packedOutput, state = self.gru(packedInput, prevState)
		padded, _ = self.padPack(packedOutput)
		logits = self.dense(padded)
		return logits, state

	def run(self):
		"""
			Training method
		"""
		self.to(self.device)
		criterion, optimizer = self.getLossAndTrainOp(0.01)

		batchCount = 0
		batchesLen = (len(self.preproObj) // self.batchSize) * self.epochs

		for e in range(self.epochs):

			batches = self.getBatches()

			stateH = self.zeroState(1, self.batchSize)[0]
			stateH = stateH.to(self.device)

			batchTimes = []

			for batch in batches:

				startTime = time.time()
				srcLengths, _ ,srcBatch,trgBatch = self.getSrcTrg(batch)

				batchCount += 1

				self.train()
				optimizer.zero_grad()

				trgBatch = torch.t(trgBatch).to(self.device)
				srcBatch = srcBatch.to(self.device)

				logits, stateH = self(srcBatch, srcLengths, stateH)

				loss = criterion(logits.transpose(1, 2), trgBatch)

				stateH = stateH.detach()

				lossValue = loss.item()

				loss.backward(retain_graph=True)

				optimizer.step()

				loss.backward()

				_ = nn.utils.clip_grad_norm_(
				self.parameters(), 5)

				optimizer.step()

				batchTime = time.time() - startTime
				batchTimes.append(batchTime)

				if batchCount % 10 == 0:
					timeChecker(batchTimes, batchCount, batchesLen, e, self.epochs, lossValue)
